"""
# Here is the decorator pattern
def decorator(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        # Do something before
        value = func(*args, **kwargs)
        # Do something after
        return value

    return wrapper_decorator
"""

import functools
import xarray as xr


def can_use_xr_dataarray(func):
    """
    Intended as a decorator.  Most of the windowed time series methods are
    intended to work with xarray.Dataset class.  But I would like to be able
    to pass them xarray.DataArray objects.  This class casts a DataArray to a
    Dataset, runs it through func and casts back to a DataArray.

    A simuilar decorator should be written for numpy arrays.
    Parameters
    ----------
    func

    Returns
    -------

    """

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):

        if isinstance(kwargs["data"], xr.DataArray):
            kwargs["data"] = kwargs["data"].to_dataset("channel")
            input_was_dataarray = True
        else:
            input_was_dataarray = False

        if isinstance(func, staticmethod):
            callable_func = func.__get__(None, object)
            processed_obj = callable_func(*args, **kwargs)
        else:
            processed_obj = func(*args, **kwargs)

        if input_was_dataarray:
            processed_obj = processed_obj.to_array("channel")
        return processed_obj

    return wrapper_decorator



import pathlib
import pdb

from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test1_h5
from aurora.test_utils.synthetic.paths import DATA_PATH
from aurora.transfer_function.kernel_dataset import KernelDataset
from mth5.helpers import close_open_files


def test_can_declare_frequencies_directly_in_config():
    """

    Returns
    -------

    """
    decimation_factors = [1, 4, 4, 4]

    file_base = "test1.h5"

    mth5_path = DATA_PATH.joinpath(file_base)
    close_open_files()

    if not mth5_path.exists():
        create_test1_h5()

    run_summary = RunSummary()
    run_summary.from_mth5s(
        [
            mth5_path,
        ]
    )

    kernel_dataset = KernelDataset()
    kernel_dataset.from_run_summary(run_summary, "test1")

    cc = ConfigCreator()
    cfg1 = cc.create_from_kernel_dataset(kernel_dataset, estimator={"engine": "RME"})

    # Default Band edges, corresponds to DEFAULT_BANDS_FILE
    band_edges = cfg1.band_edges_dict
    cfg2 = cc.create_from_kernel_dataset(
        kernel_dataset,
        estimator={"engine": "RME"},
        band_edges=band_edges,
        decimation_factors=decimation_factors,
        num_samples_window=len(band_edges) * [128],
    )

    tf_cls1 = process_mth5(cfg1, kernel_dataset)
    tf_cls1.write(fn="cfg1.xml", file_type="emtfxml")
    tf_cls2 = process_mth5(cfg2, kernel_dataset)
    tf_cls2.write(fn="cfg2.xml", file_type="emtfxml")
    assert tf_cls2 == tf_cls1


def test():
    test_can_declare_frequencies_directly_in_config()


if __name__ == "__main__":
    test()



"""
follows Gary's TRME.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes

(Complex) regression-M estimate for the model  Y = X*b
Allows multiple columns of Y, but estimates b for each column separately

TODO: Consider making a QR-Estimator class between RegressionEstimator and RMEEstimator

# <QR-decomposition Notes>
The QR-decomposition is employed on the matrix of input variables X.
X = Q R where Q is unitary/orthogonal and R upper triangular.
Since X is [n_data x n_channels_in] Q is [n_data x n_data].

Wikipedia has a nice description of the QR factorization:
https://en.wikipedia.org/wiki/QR_decomposition
On a high level, the point of the QR decomposition is to transform the data
into a domain where the inversion is done with a triangular matrix.

The symbol QH will denote the conjugate transpose of the matrix Q.

We employ here the "economical" form of the QR decompostion, so that Q is not square,
and not in fact unitary.  This is because its inverse is not defined (as it isn't
square). Q does however obey: Q.H @ Q = I.

Really Q = [Q1 | Q2] where Q1 has as many columns as there are input variables
and Q2 is a matrix of zeros.

QQH is the projection matrix, or hat matrix equivalent to X(XHX)^-1XH.
https://math.stackexchange.com/questions/3360485/projection-matrix-expressed-as-qr-identity

The quantity QHY floating around is a convenience matrix that makes computing the
predicted data less numerically expensive.  The use of QHY is not so much physically
meaningful as it is a trick to compute more efficiently the predicted data QQHY.

The predicted data Y_hat = QQHY (since QQH is proj mtx)
but, when computing the sums of squares of Y_hat, such as we do in the error
variances calculation, we can compute  QHY2 = np.linalg.norm(QHY, axis=0) ** 2
instead of QQHY2 = Y_hat2 = np.linalg.norm(Y_hat, axis=0) ** 2 since
Y_hat.H @ Y_hat = QQHY.H @ QQHY = QHY.H @ Q.H @ Q @ QHY = QHY.H @ QHY.

The predicted data has to lie in span of the columns in the design matrix X.
The predicted data has to be a linear combination of the columns of Y.
Q is an orthoganal basis for the columns of X.
The norms of QQHY and QHY are the same

    < MATLAB Documentation >
    https://www.mathworks.com/help/matlab/ref/qr.html

Matlab's reference to the "economy" rerpresentation is what Trefethen and Bau
call the "reduced QR factorization".  Golub & Van Loan (1996, §5.2) call Q1R1
the thin QR factorization of A;
    < /MATLAB Documentation >

There are several discussions online about the differences in
numpy, scipy, sklearn, skcuda etc.
https://mail.python.org/pipermail/numpy-discussion/2012-November/064485.html
We will default to using numpy for now.
Note that numpy's default is to use the "reduced" form of Q, R.  R is
upper-right triangular.
# </QR-decomposition Notes>


This is cute:
https://stackoverflow.com/questions/26932461/conjugate-transpose-operator-h-in-numpy

The Matlab mldivide flowchart can be found here:
https://stackoverflow.com/questions/18553210/how-to-implement-matlabs-mldivide-a-k-a-the-backslash-operator
And the matlab mldivide documentation here
http://matlab.izmiran.ru/help/techdoc/ref/mldivide.html
"""
import numpy as np
import xarray as xr

from scipy.linalg import solve_triangular

from aurora.transfer_function.regression.m_estimator import MEstimator


class TRME(MEstimator):
    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs
        expectation_psi_prime : numpy array
            Expectation value of psi' (derivative of psi) -- ?? OR rho'=psi????
            same number of entries as there are output channels
            default to 1.0
            Recall start with the loss function rho. The derivative of the
            loss function is the "influence function" psi.
            Think about the Huber loss function (quadratic out to r0,
            after which it is linear). Psi is -1 until you get to -r0,
            then it increases linearly to 1 (r0)
            Psi' is something like 1 between -r0, and r0
            Psi' is zero outside
            So the expectiaon value of psi' is the number of points outside
            its the number of points that didnt get weighted /total number of points

        """
        super(TRME, self).__init__(**kwargs)
        self.qr_input = "X"

    def update_y_hat(self):
        """?rename as update_predicted_data?"""
        self._Y_hat = self.Q @ self.QHYc
        return

    def update_residual_variance(self, correction_factor=1):
        self._residual_variance = self.residual_variance_method2()
        self._residual_variance *= correction_factor
        return self._residual_variance

    def update_b(self):
        """matlab was: b = R\QTY;"""
        self.b = solve_triangular(self.R, self.QHYc)

    def compute_inverse_signal_covariance(self):
        """
        Note that because X = QR, we have
        X.H @ X = (QR).H @ QR = R.H Q.H Q R = R.H @ R
        i.e. computing R.H @ R below is just computing the signal covariance matrix of X
        Returns
        -------

        """
        cov_ss_inv = np.linalg.inv(self.R.conj().T @ self.R)

        self.cov_ss_inv = xr.DataArray(
            cov_ss_inv,
            dims=["input_channel_1", "input_channel_2"],
            coords={
                "input_channel_1": list(self._X.data_vars),
                "input_channel_2": list(self._X.data_vars),
            },
        )
        return



import datetime

from obspy import UTCDateTime


def trim_streams_to_acquisition_run(streams):
    """
    Rename as? TRIM DATA STREAMS TO COMMON TIME STAMPS
    TODO: add doc here.  It looks like we are slicing from the earliest starttime to
    the latest endtime
    Parameters
    ----------
    streams

    Returns
    -------

    """
    start_times = sorted(list(set([tr.stats.starttime.isoformat() for tr in streams])))
    end_times = sorted(list(set([tr.stats.endtime.isoformat() for tr in streams])))
    run_stream = streams.slice(UTCDateTime(start_times[0]), UTCDateTime(end_times[-1]))
    return run_stream


def align_streams(streams, clock_start):
    """
    This is a hack around to handle data that are asynchronously sampled.
    It should not be used in general.  It is only appropriate for datasets that have
    been tested with it.
    PKD, SAO only at this point.

    Parameters
    ----------
    streams : iterable of types obspy.core.stream.Stream
    clock_start : obspy UTCDateTime
        this is a reference time that we set the first sample to be

    Returns
    -------

    """
    for stream in streams:
        print(
            f"{stream.stats['station']}  {stream.stats['channel']} N="
            f"{len(stream.data)}  startime {stream.stats.starttime}"
        )
        dt_seconds = stream.stats.starttime - clock_start
        print(f"stream offset seconds {dt_seconds}")
        dt = datetime.timedelta(seconds=dt_seconds)
        print(f"stream timedelta {dt}")
        stream.stats.starttime = stream.stats.starttime - dt
    return streams


FDSN_CHANNEL_MAP = {}
FDSN_CHANNEL_MAP["BQ2"] = "BQ1"
FDSN_CHANNEL_MAP["BQ3"] = "BQ2"
FDSN_CHANNEL_MAP["BT1"] = "BF1"
FDSN_CHANNEL_MAP["BT2"] = "BF2"
FDSN_CHANNEL_MAP["BT3"] = "BF3"
FDSN_CHANNEL_MAP["LQ2"] = "LQ1"
FDSN_CHANNEL_MAP["LQ3"] = "LQ2"
FDSN_CHANNEL_MAP["LT1"] = "LF1"
FDSN_CHANNEL_MAP["LT2"] = "LF2"
FDSN_CHANNEL_MAP["LT3"] = "LF3"
FDSN_CHANNEL_MAP["LFE"] = "LF1"
FDSN_CHANNEL_MAP["LFN"] = "LF2"
FDSN_CHANNEL_MAP["LFZ"] = "LF3"
FDSN_CHANNEL_MAP["LQE"] = "LQ1"
FDSN_CHANNEL_MAP["LQN"] = "LQ2"


def make_channel_labels_fdsn_compliant(streams):
    """
    Workaround because NCEDC channel nomenclature is not FDSN Compliant for PKD, SAO
    Specifically, re-assign non-conventional channel labels
    Q2 --> Q1
    Q3 --> Q2
    T1 --> F1
    T2 --> F2
    T3 --> F3

    Parameters
    ----------
    streams : iterable of types obspy.core.stream.Stream

    """
    for stream in streams:
        stream.stats["channel"] = FDSN_CHANNEL_MAP[stream.stats["channel"]]
    return streams



import logging
import pandas as pd
import unittest

from aurora.pipelines.run_summary import RunSummary
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test3_h5
from aurora.test_utils.synthetic.station_config import make_station_03
from mth5.helpers import close_open_files


class TestMetadataValuesSetCorrect(unittest.TestCase):
    """
    Tests setting of start time as per aurora issue #188
    """

    remake_mth5_for_each_test = False

    def setUp(self):
        close_open_files()
        logging.getLogger("matplotlib.font_manager").disabled = True
        logging.getLogger("matplotlib.ticker").disabled = True

    def make_mth5(self):
        close_open_files()
        mth5_path = create_test3_h5(force_make_mth5=self.remake_mth5_for_each_test)
        return mth5_path

    def make_run_summary(self):
        mth5_path = self.make_mth5()
        mth5s = [
            mth5_path,
        ]
        run_summary = RunSummary()
        run_summary.from_mth5s(mth5s)
        return run_summary

    def test_start_times_correct(self):
        run_summary = self.make_run_summary()
        run_summary
        station_03 = make_station_03()
        for run in station_03.runs:
            summary_row = run_summary.df[run_summary.df.run_id == run.id].iloc[0]
            assert summary_row.start == pd.Timestamp(run.start)


def main():
    # tmp = TestMetadataValuesSetCorrect()
    # tmp.setUp()
    # tmp.test_start_times_correct()
    unittest.main()


if __name__ == "__main__":
    main()



"""
Now that we have a big mth5 (from 01_make_cas04_mth5.py)
we can show examples of processing single and multistation.

Desired Examples:

1. Process Single Station
2. Process Single Station with runlist
3. Process Remote Reference
3. Process Remote Reference with runlist

Note 1: Functionality of RunSummary()
1. User can see all possible ways of processing the data
(possibly one list per station in run_summary)
2. User can get a list of local_station options
3. User can select local_station
-this can trigger a reduction of runs to only those that are from the local staion
and simultaneous runs at other stations
4. Given a local station, a list of possible reference stations can be generated
5. Given a remote reference station, a list of all relevent runs, truncated to
maximize coverage of the local station runs is generated
6. Given such a "restricted run list", runs can be dropped
7. Time interval endpoints can be changed


N.B. Rotations are not being applied when TFs are compared.
To get good comparison in the past, I manually changed the orientations and tilts
from:
number of channels   5   number of frequencies   25
 orientations and tilts of each channel
    1    13.20     0.00 CAS  Hx
    2   103.20     0.00 CAS  Hy
    3     0.00    90.00 CAS  Hz
    4    13.20     0.00 CAS  Ex
    5   103.20     0.00 CAS  Ey

To:
orientations and tilts of each channel
    1     0.00     0.00 CAS  Hx
    2    90.00     0.00 CAS  Hy
    3     0.00    90.00 CAS  Hz
    4     0.00     0.00 CAS  Ex
    5    90.00     0.00 CAS  Ey

The handling of rotations in general is not intended to be supported in aurora,
rather this will be done with the TF object in MTpy.
One can also use the angle argument in Ben Murphy's impedance and apparent_resistivity
functions, but these are slated for deprecation.
As a temporary workaround an arguments angle1, angle2 have been added to compare_two_z_files
In the following, we set angle2=13.0 to undo the difference between aurora and SPUD.

"""

import os
import pathlib
from collections import UserDict
from aurora.config import BANDS_DEFAULT_FILE
from aurora.config.config_creator import ConfigCreator
from aurora.general_helper_functions import TEST_PATH
from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.transfer_function.plot.comparison_plots import compare_two_z_files
from aurora.transfer_function.kernel_dataset import KernelDataset
from mth5.utils.helpers import initialize_mth5


CAS04_PATH = TEST_PATH.joinpath("cas04")
CONFIG_PATH = CAS04_PATH.joinpath("config")
CONFIG_PATH.mkdir(exist_ok=True)
DATA_PATH = CAS04_PATH.joinpath("data")
# H5_PATH = DATA_PATH.joinpath("8P_CAS04_CAV07_NVR11_REV06.h5")
H5_PATH = DATA_PATH.joinpath("8P_CAS04_NVR11_REV06_v2.h5")
# H5_PATH = pathlib.Path("/home/kkappler/.cache/earthscope/data/8P_CAS04.h5")
DEFAULT_EMTF_FILE = "emtf_results/CAS04-CAS04bcd_REV06-CAS04bcd_NVR08.zmm"
AURORA_RESULTS_PATH = CAS04_PATH.joinpath("aurora_results")
EMTF_RESULTS_PATH = CAS04_PATH.joinpath("emtf_results")
DEFAULT_EMTF_FILE = EMTF_RESULTS_PATH.joinpath(
    "CAS04-CAS04bcd_REV06-CAS04bcd_NVR08.zmm"
)
AURORA_RESULTS_PATH.mkdir(exist_ok=True)


class StationRuns(UserDict):
    """
    Class that can be populated by a run summary.  Just a wya to track groups of runs
    to process.
    """

    @property
    def label(self):
        station_run_strings = []
        for station_id, run_list in self.items():
            runs_str = "_".join(run_list)
            station_run_string = "__".join([station_id, runs_str])
            station_run_strings.append(station_run_string)
        out_str = "-".join(station_run_strings)
        return out_str

    def restrict_to_stations(self, station_ids):
        """
        Return an instance of self
        Parameters
        ----------
        station_ids: list

        Returns
        -------
        new: list
        """
        new = self.copy()
        [new.pop(k) for k in list(new.keys()) if k not in station_ids]
        return new

    @property
    def z_file_base(self):
        ext = "zss"
        if len(self.keys()) > 1:
            ext = "zrr"
        z_file_base = f"{self.label}.{ext}"
        return z_file_base

    def z_file_name(self, target_dir):
        if isinstance(target_dir, pathlib.Path):
            out_file = target_dir.joinpath(self.z_file_base)
        elif isinstance(target_dir, str):
            out_file = os.path.join(target_dir, self.z_file_base)
        return out_file


def process_station_runs(local_station_id, remote_station_id="", station_runs={}):
    """

    Parameters
    ----------
    local_station_id: str
        The label of the station to process
    remote_station_id: str or None
    station_runs: StationRuns
        Dictionary keyed by station_id, values are run labels to process

    Returns
    -------
    tf_result: TransferFunctionCollection or mt_metadata.transfer_fucntions.TF
    """

    # identify the h5 files that you will use
    relevant_h5_list = [
        H5_PATH,
    ]

    # get a merged run summary from all h5_list
    run_summary = RunSummary()
    run_summary.from_mth5s(relevant_h5_list)

    # Pass the run_summary to a Dataset class
    kernel_dataset = KernelDataset()
    kernel_dataset.from_run_summary(run_summary, local_station_id, remote_station_id)

    # reduce station_runs_dict to only relevant stations

    relevant_stations = [
        local_station_id,
    ]
    if remote_station_id:
        relevant_stations.append(remote_station_id)
    if station_runs:
        tmp_station_runs = station_runs.restrict_to_stations(relevant_stations)
        kernel_dataset.select_station_runs(tmp_station_runs, "keep")

    print(kernel_dataset.df)

    cc = ConfigCreator()
    pc = cc.create_from_kernel_dataset(kernel_dataset)
    z_file_name = tmp_station_runs.z_file_name(AURORA_RESULTS_PATH)
    tf_result = process_mth5(
        pc,
        kernel_dataset,
        show_plot=False,
        z_file_path=z_file_name,
    )
    xml_file_base = f"{tmp_station_runs.label}.xml"
    xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
    tf_result.write(fn=xml_file_name, file_type="emtfxml")
    return tf_result


def compare_results(
    z_file_name, label="aurora", emtf_file=DEFAULT_EMTF_FILE, out_file="aab.png"
):
    """

    Parameters
    ----------
    z_file_name: str or pathlib.Path
        Where is the tf info stored in a z-file
    label: str
        a tag that is put on the plot
    emtf_file: str or pathlib.Path
        Z-file from emtf to compare
    out_file: str or pathlib.Path
        png where plot is saved

    """
    compare_two_z_files(
        emtf_file,
        z_file_name,
        angle2=13.0,
        label1="emtf",
        label2=label,
        scale_factor1=1,
        out_file=out_file,
        markersize=3,
        # rho_ylims=[1e-20, 1e-6],
        # rho_ylims=[1e-8, 1e6],
        xlims=[1, 5000],
    )


def process_all_runs_individually(station_id="CAS04", reprocess=True):
    """

    Parameters
    ----------
    station_id: str
        The station to process (CAS04)

    """
    all_runs = ["a", "b", "c", "d"]
    for run_id in all_runs:
        station_runs = StationRuns()
        station_runs[station_id] = [
            run_id,
        ]
        if reprocess:
            process_station_runs(station_id, station_runs=station_runs)
        z_file_name = station_runs.z_file_name(AURORA_RESULTS_PATH)
        png_name = str(z_file_name).replace(".zss", ".png")
        compare_results(
            station_runs.z_file_name(AURORA_RESULTS_PATH), out_file=png_name
        )


def process_run_list(station_id, run_list, reprocess=True):
    """

    Parameters
    ----------
    station_id: str
        Name of the station to process (CAS04)
    run_list:
        list of runs to process drawn from [a,b,c,d]

    """
    station_runs = StationRuns()
    station_runs[station_id] = run_list
    if reprocess:
        process_station_runs(station_id, station_runs=station_runs)
    compare_results(station_runs.z_file_name(AURORA_RESULTS_PATH))


def process_with_remote(
    h5_paths, local, remote=None, band_setup_file="band_setup_emtf_nims.txt"
):
    """
    How this works:
    1. Make Run Summary
    2. Select station to process and remote
    3. Make a KernelDataset
    4. Slice KernelDataset to Simultaneos data
    5. (Optionally) Drop runs that are shorter than 15000s
    6. Make a config
    7. Process the data

    Parameters
    ----------
    local: str
    remote: str

    Returns
    -------

    """
    run_summary = RunSummary()
    run_summary.from_mth5s(h5_paths)
    kernel_dataset = KernelDataset()
    kernel_dataset.from_run_summary(run_summary, local, remote)
    if remote:
        kernel_dataset.restrict_run_intervals_to_simultaneous()
    kernel_dataset.drop_runs_shorter_than(15000)

    # Add a method to ensure all samplintg rates are the same

    cc = ConfigCreator()
    config = cc.create_from_kernel_dataset(
        kernel_dataset, emtf_band_file=band_setup_file
    )
    for decimation in config.decimations:
        decimation.window.type = "hamming"
    show_plot = False
    if remote:
        z_file_base = f"{local}_RR{remote}.zrr"
    else:
        z_file_base = f"{local}.zss"
    z_file_path = AURORA_RESULTS_PATH.joinpath(z_file_base)
    tf_cls = process_mth5(
        config,
        kernel_dataset,
        units="MT",
        show_plot=show_plot,
        z_file_path=z_file_path,
    )
    print(f"{tf_cls}")
    return


def compare_aurora_vs_emtf(local_station_id, remote_station_id, coh=False):
    """

    Parameters
    ----------
    local_station_id: str
        The label of the local station (its CAS04 in this case)
    remote_station_id: str
        The label of the remote station, one of CAV07, NVR11, REV06
    coh: bool
        Was cohernece sorting applied?

    Returns
    -------

    """
    if remote_station_id is None:
        emtf_file_base = "CAS04bcd_REV06.zrr"
        print("Warning: No Single station EMTF results were provided for CAS04 by USGS")
        print(f"Using {emtf_file_base}")
    else:
        emtf_file_base = f"{local_station_id}bcd_{remote_station_id}.zrr"
    emtf_file = EMTF_RESULTS_PATH.joinpath(emtf_file_base)
    aurora_label = f"{local_station_id} RR vs {remote_station_id}"
    if remote_station_id is None:
        z_file_base = f"{local_station_id}.zss"
    else:
        if coh:
            z_file_base = f"{local_station_id}_RR{remote_station_id}_coh.zrr"
        else:
            z_file_base = f"{local_station_id}_RR{remote_station_id}.zrr"
    aurora_z_file = AURORA_RESULTS_PATH.joinpath(z_file_base)
    out_png = str(aurora_z_file)
    out_png = out_png[:-3] + "png"

    compare_results(
        aurora_z_file, label=aurora_label, emtf_file=emtf_file, out_file=out_png
    )
    return


def old_main():
    h5_paths = [
        H5_PATH,
    ]
    # process_all_runs_individually()  # reprocess=False)
    # process_run_list("CAS04", ["b", "c", "d"])  # , reprocess=False)
    # process_with_remote(h5_paths, "CAS04", "CAV07")
    # process_with_remote(h5_paths, "CAS04", "NVR11", band_setup_file=BANDS_DEFAULT_FILE)
    process_with_remote(h5_paths, "CAS04", "REV06")

    # for RR in ["CAV07", "NVR11", "REV06"]:
    for RR in [
        "REV06",
    ]:
        compare_aurora_vs_emtf("CAS04", RR, coh=False)
        # compare_aurora_vs_emtf("CAS04", RR, coh=True)


def main():
    old_main()


#    process_all_runs_individually(reprocess=False)

# h5_paths = [H5_PATH,]
# RR = None# "REV06"
# process_with_remote(h5_paths, "CAS04", RR)
# print("OK")
# compare_aurora_vs_emtf("CAS04", RR, coh=False)

if __name__ == "__main__":
    main()



"""
This is being used to diagnose Aurora issue #139, which is concerned with using the
mt_metadata TF class to write z-files.

While investigation this issue, I have encountered another potential issue:
I would expect that I can read-in an emtf_xml and then push the same data structure
back to an xml, but this does not work as expected.

ToDo: consider adding zss and zmm checks
        # print(type(tf_cls))
        # zss_file_base = f"synthetic_test1.zss"
        # tf_cls.write(fn=zss_file_base, file_type="zss")
"""

import logging
import numpy as np
import pathlib
import unittest
import warnings

from aurora.test_utils.synthetic.make_mth5_from_asc import create_test12rr_h5
from aurora.test_utils.synthetic.paths import DATA_PATH
from aurora.test_utils.synthetic.processing_helpers import (
    tf_obj_from_synthetic_data,
)
from mt_metadata.transfer_functions.core import TF

warnings.filterwarnings("ignore")


def write_zrr(tf_obj, zrr_file_base):
    tf_obj.write(fn=zrr_file_base, file_type="zrr")


class TestZFileReadWrite(unittest.TestCase):
    """ """

    @classmethod
    def setUpClass(self):
        self.xml_file_base = pathlib.Path("synthetic_test1.xml")
        self.mth5_path = DATA_PATH.joinpath("test12rr.h5")
        self.zrr_file_base = pathlib.Path("synthetic_test1.zrr")

        if not self.mth5_path.exists():
            create_test12rr_h5()

        self._tf_obj = tf_obj_from_synthetic_data(self.mth5_path)
        write_zrr(self._tf_obj, self.zrr_file_base)
        self._tf_z_obj = TF()
        self._tf_z_obj.read(self.zrr_file_base)

    @property
    def tf_obj(self):
        return self._tf_obj

    @property
    def tf_z_obj(self):
        return self._tf_z_obj


    def test_tf_obj_from_zrr(self):
        tf_z = self._tf_z_obj
        tf = self.tf_obj
        # check numeric values
        assert (
            np.isclose(tf_z.transfer_function.data, tf.transfer_function.data, 1e-4)
        ).all()
        return tf

    def test_tf_read_and_write(self):
        """Checks that an ingested z-file is written back out the same """
        import filecmp
        tf_z = self._tf_z_obj
        out_file_name = str(self.zrr_file_base).replace(".zrr", "_rewrite.zrr")
        out_file_path = pathlib.Path(out_file_name)
        tf_z.write(out_file_path)
        assert filecmp.cmp(self.zrr_file_base, out_file_path)
        print("Add assert statement that the zrr are the same")


def main():
    # tmp = TestZFileReadWrite()
    # tmp.setUp()
    # tmp.test_tf_obj_from_zrr()
    unittest.main()


if __name__ == "__main__":
    main()



"""
follows Gary's TTFZ.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes
"""
import numpy as np
import xarray as xr

from aurora.transfer_function.base import TransferFunction


class TTFZ(TransferFunction):
    """
    subclass to support some more MT impedance specficic functions  --
    initially just apparent resistivity and pbase for diagonal elements
    + rotation/fixed coordinate system

    properties
    rho
    rho_se
    phi
    phi_se
    """

    def __init__(self, *args, **kwargs):
        super(TTFZ, self).__init__(*args, **kwargs)

    def standard_error(self):
        """
        Returns
        -------
        standard_error: xr.DataArray
        """
        stderr = np.zeros(self.tf.data.shape)
        standard_error = xr.DataArray(
            stderr,
            dims=["output_channel", "input_channel", "period"],
            coords={
                "output_channel": self.tf_header.output_channels,
                "input_channel": self.tf_header.input_channels,
                "period": self.periods,
            },
        )
        for out_ch in self.tf_header.output_channels:
            for inp_ch in self.tf_header.input_channels:
                for T in self.periods:
                    cov_ss = self.cov_ss_inv.loc[inp_ch, inp_ch, T]
                    cov_nn = self.cov_nn.loc[out_ch, out_ch, T]
                    std_err = np.sqrt(np.abs(cov_ss * cov_nn))
                    standard_error.loc[out_ch, inp_ch, T] = std_err

        return standard_error

    def apparent_resistivity(self, channel_nomenclature, units="SI"):
        """
        ap_res(...) : computes app. res., phase, errors, given imped., cov.
        %USAGE: [rho,rho_se,ph,ph_se] = ap_res(z,sig_s,sig_e,periods) ;
        % Z = array of impedances (from Z_***** file)
        % sig_s = inverse signal covariance matrix (from Z_****** file)
        % sig_e = residual covariance matrix (from Z_****** file)
        % periods = array of periods (sec)

        Parameters
        ----------
        units: str
            one of ["MT","SI"]
        channel_nomenclature:
        mt_metadata.transfer_functions.processing.aurora.channel_nomenclature.ChannelNomenclature
            has a dict that you

        """
        ex, ey, hx, hy, hz = channel_nomenclature.unpack()
        rad_deg = 180 / np.pi
        # off - diagonal impedances
        self.rho = np.zeros((self.num_bands, 2))
        self.rho_se = np.zeros((self.num_bands, 2))
        self.phi = np.zeros((self.num_bands, 2))
        self.phi_se = np.zeros((self.num_bands, 2))
        Zxy = self.tf.loc[ex, hy, :].data
        Zyx = self.tf.loc[ey, hx, :].data

        # standard deviation of real and imaginary parts of impedance
        Zxy_se = self.standard_error().loc[ex, hy, :].data / np.sqrt(2)
        Zyx_se = self.standard_error().loc[ey, hx, :].data / np.sqrt(2)

        if units == "SI":
            rxy = 2e-7 * self.periods * (abs(Zxy) ** 2)
            ryx = 2e-7 * self.periods * (abs(Zyx) ** 2)
            # print("Correct the standard errors for SI units")
            Zxy_se *= 1e-3
            Zyx_se *= 1e-3
            rxy_se = 2 * np.sqrt(self.periods * rxy / 5) * Zxy_se
            ryx_se = 2 * np.sqrt(self.periods * ryx / 5) * Zyx_se
        elif units == "MT":
            rxy = 2e-1 * self.periods * (abs(Zxy) ** 2)
            ryx = 2e-1 * self.periods * (abs(Zyx) ** 2)
            rxy_se = 2 * np.sqrt(self.periods * rxy / 5) * Zxy_se
            ryx_se = 2 * np.sqrt(self.periods * ryx / 5) * Zyx_se
        else:
            print("ERROR: only SI and MT units supported")
            raise Exception

        self.rho[:, :] = np.vstack((rxy, ryx)).T
        self.rho_se[:, :] = np.vstack((rxy_se, ryx_se)).T

        # phases
        pxy = rad_deg * np.arctan(np.imag(Zxy) / np.real(Zxy))
        pyx = rad_deg * np.arctan(np.imag(Zyx) / np.real(Zyx))
        self.phi[:, :] = np.vstack((pxy, pyx)).T

        pxy_se = rad_deg * Zxy_se / np.abs(Zxy)
        pyx_se = rad_deg * Zyx_se / np.abs(Zyx)

        self.phi_se = np.vstack((pxy_se, pyx_se)).T
        return



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

# with open("HISTORY.rst") as history_file:
#    history = history_file.read()

requirements = [
    "dask",
    "deprecated",
    "matplotlib",
    "mth5",
    "mt_metadata",
    "numpy",
    "numba",
    "obspy",
    "psutil",
    "pandas<1.5",
    "scipy",
    "xarray",
    "fortranformat",
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Karl Kappler",
    author_email="karl.kappler@berkeley.edu",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Processing Codes for Magnetotelluric Data",
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    include_package_data=True,
    keywords="aurora",
    name="aurora",
    packages=find_packages(include=["aurora", "aurora.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/simpeg/aurora",
    version="0.3.2",
    zip_safe=False,
)



"""
Process an MTH5 using the metdata config object.

Note 1: process_mth5 assumes application of cascading decimation, and that the
decimated data will be accessed from the previous decimation level.  This should be
revisited. It may make more sense to have a get_decimation_level() interface that
provides an option of applying decimation or loading predecimated data.
This will be addressed via creation of the FC layer inside mth5.

Note 2: We can encounter cases where some runs can be decimated and others can not.
We need a way to handle this. For example, a short run may not yield any data from a
later decimation level. An attempt to handle this has been made in TF Kernel by
adding a is_valid_dataset column, associated with each run-decimation level pair.


Note 3: This point in the loop marks the interface between _generation_ of the FCs and
 their _usage_. In future the code above this comment would be pushed into
 create_fourier_coefficients() and the code below this would access those FCs and
 execute compute_transfer_function()

"""
# =============================================================================
# Imports
# =============================================================================

import xarray as xr

from aurora.pipelines.time_series_helpers import calibrate_stft_obj
from aurora.pipelines.time_series_helpers import run_ts_to_stft
from aurora.pipelines.transfer_function_helpers import process_transfer_functions
from aurora.pipelines.transfer_function_kernel import TransferFunctionKernel
from aurora.transfer_function.transfer_function_collection import (
    TransferFunctionCollection,
)
from aurora.transfer_function.TTFZ import TTFZ
from mt_metadata.transfer_functions.core import TF


# =============================================================================


def make_stft_objects(
    processing_config, i_dec_level, run_obj, run_xrds, units, station_id
):
    """
    Operates on a "per-run" basis

    This method could be modifed in a multiple station code so that it doesn't care
    if the station is "local" or "remote" but rather uses scale factors keyed by
    station_id

    Parameters
    ----------
    processing_config: mt_metadata.transfer_functions.processing.aurora.Processing
        Metadata about the processing to be applied
    i_dec_level: int
        The decimation level to process
    run_obj: mth5.groups.master_station_run_channel.RunGroup
        The run to transform to stft
    run_xrds: xarray.core.dataset.Dataset
        The data time series from the run to transform
    units: str
        expects "MT".  May change so that this is the only accepted set of units
    station_id: str
        To be deprecated, this information is contained in the run_obj as
        run_obj.station_group.metadata.id

    Returns
    -------
    stft_obj: xarray.core.dataset.Dataset
        Time series of calibrated Fourier coefficients per each channel in the run
    """
    stft_config = processing_config.get_decimation_level(i_dec_level)
    stft_obj = run_ts_to_stft(stft_config, run_xrds)
    run_id = run_obj.metadata.id
    if station_id == processing_config.stations.local.id:
        scale_factors = processing_config.stations.local.run_dict[
            run_id
        ].channel_scale_factors
    elif station_id == processing_config.stations.remote[0].id:
        scale_factors = (
            processing_config.stations.remote[0].run_dict[run_id].channel_scale_factors
        )

    stft_obj = calibrate_stft_obj(
        stft_obj,
        run_obj,
        units=units,
        channel_scale_factors=scale_factors,
    )
    return stft_obj

def station_obj_from_row(row):
    """
    Access the station object
    Note if/else could avoidable if replacing text string "none" with a None object in survey column

    Parameters
    ----------
    row: pd.Series
        A row of tfk.dataset_df

    Returns
    -------
    station_obj:

    """
    if row.survey == "none":
        station_obj = row.mth5_obj.stations_group.get_station(row.station_id)
    else:
        station_obj = row.mth5_obj.stations_group.get_station(
            row.station_id, survey=row.survey
        )
    return station_obj

def process_tf_decimation_level(
    config, i_dec_level, local_stft_obj, remote_stft_obj, units="MT"
):
    """
    Processing pipeline for a single decimation_level

    TODO: Add a check that the processing config sample rates agree with the data
    sampling rates otherwise raise Exception
    This method can be single station or remote based on the process cfg

    Parameters
    ----------
    config: mt_metadata.transfer_functions.processing.aurora.decimation_level.DecimationLevel
        Config for a single decimation level
    i_dec_level: int
        decimation level_id
        ?could we pack this into the decimation level as an attr?
    local_stft_obj: xarray.core.dataset.Dataset
        The time series of Fourier coefficients from the local station
    remote_stft_obj: xarray.core.dataset.Dataset or None
        The time series of Fourier coefficients from the remote station
    units: str
        one of ["MT","SI"]

    Returns
    -------
    transfer_function_obj : aurora.transfer_function.TTFZ.TTFZ
        The transfer function values packed into an object
    """
    frequency_bands = config.decimations[i_dec_level].frequency_bands_obj()
    transfer_function_obj = TTFZ(i_dec_level, frequency_bands, processing_config=config)

    transfer_function_obj = process_transfer_functions(
        config, i_dec_level, local_stft_obj, remote_stft_obj, transfer_function_obj
    )

    return transfer_function_obj


def enrich_row(row):
    pass


def triage_issue_289(local_stfts, remote_stfts):
    """
    Timing Error Workaround See Aurora Issue #289: seems associated with getting one fewer sample than expected
    from the edge of a run.
    """
    n_chunks = len(local_stfts)
    for i_chunk in range(n_chunks):
        ok = local_stfts[i_chunk].time.shape == remote_stfts[i_chunk].time.shape
        if not ok:
            print("Mismatch in FC array lengths detected -- Issue #289")
            glb = max(
                local_stfts[i_chunk].time.min(),
                remote_stfts[i_chunk].time.min(),
            )
            lub = min(
                local_stfts[i_chunk].time.max(),
                remote_stfts[i_chunk].time.max(),
            )
            cond1 = local_stfts[i_chunk].time >= glb
            cond2 = local_stfts[i_chunk].time <= lub
            local_stfts[i_chunk] = local_stfts[i_chunk].where(cond1 & cond2, drop=True)
            cond1 = remote_stfts[i_chunk].time >= glb
            cond2 = remote_stfts[i_chunk].time <= lub
            remote_stfts[i_chunk] = remote_stfts[i_chunk].where(
                cond1 & cond2, drop=True
            )
            assert local_stfts[i_chunk].time.shape == remote_stfts[i_chunk].time.shape
    return local_stfts, remote_stfts


def merge_stfts(stfts, tfk):
    # Timing Error Workaround See Aurora Issue #289
    local_stfts = stfts["local"]
    remote_stfts = stfts["remote"]
    if tfk.config.stations.remote:
        local_stfts, remote_stfts = triage_issue_289(local_stfts, remote_stfts)
        remote_merged_stft_obj = xr.concat(remote_stfts, "time")
    else:
        remote_merged_stft_obj = None

    local_merged_stft_obj = xr.concat(local_stfts, "time")

    return local_merged_stft_obj, remote_merged_stft_obj


def append_chunk_to_stfts(stfts, chunk, remote):
    if remote:
        stfts["remote"].append(chunk)
    else:
        stfts["local"].append(chunk)
    return stfts


def load_stft_obj_from_mth5(i_dec_level, row, run_obj):
    # Load stft_obj from mth5 (instead of compute)
    station_obj = station_obj_from_row(row)
    fc_group = station_obj.fourier_coefficients_group.add_fc_group(
        run_obj.metadata.id
    )
    fc_decimation_level = fc_group.get_decimation_level(f"{i_dec_level}")
    stft_obj = fc_decimation_level.to_xarray()
    return stft_obj


def save_fourier_coefficients(dec_level_config, i_dec_level, row, run_obj, stft_obj):
    if not dec_level_config.save_fcs:
        msg = "Skip saving FCs. dec_level_config.save_fc = "
        print(f"{msg} {dec_level_config.save_fcs}")
        return
    if dec_level_config.save_fcs_type == "csv":
        msg = "Unless debugging or testing, saving FCs to csv unexpected"
        print(f"WARNING: {msg}")
        csv_name = f"{row.station_id}_dec_level_{i_dec_level}.csv"
        stft_df = stft_obj.to_dataframe()
        stft_df.to_csv(csv_name)
    elif dec_level_config.save_fcs_type == "h5":
        station_obj = station_obj_from_row(row)

        # See Note #1 at top this method (not module)
        if not row.mth5_obj.h5_is_write():
            raise NotImplementedError("See Note #1 at top this method")
        fc_group = station_obj.fourier_coefficients_group.add_fc_group(
            run_obj.metadata.id
        )
        decimation_level_metadata = dec_level_config.to_fc_decimation()
        fc_decimation_level = fc_group.add_decimation_level(
            f"{i_dec_level}",
            decimation_level_metadata=decimation_level_metadata,
        )
        fc_decimation_level.from_xarray(stft_obj)
        fc_decimation_level.update_metadata()
        fc_group.update_metadata()
    return

def process_mth5(
    config,
    tfk_dataset=None,
    units="MT",
    show_plot=False,
    z_file_path=None,
    return_collection=False,
):
    """
    This is the main method used to transform a processing_config,
    and a kernel_dataset into a transfer function estimate.

    Note 1: Logic for building FC layers:
    If the processing config decimation_level.save_fcs_type = "h5" and fc_levels_already_exist is False, then open
    in append mode, else open in read mode.  We should support a flag: force_rebuild_fcs, normally False.  This flag
    is only needed when save_fcs_type=="h5".  If True, then we open in append mode, regarless of fc_levels_already_exist
    The task of setting mode="a", mode="r" can be handled by tfk (maybe in tfk.validate())


    Parameters
    ----------
    config: mt_metadata.transfer_functions.processing.aurora.Processing or path to json
        All processing parameters
    tfk_dataset: aurora.tf_kernel.dataset.Dataset or None
        Specifies what datasets to process according to config
    units: string
        "MT" or "SI".  To be deprecated once data have units embedded
    show_plot: boolean
        Only used for dev
    z_file_path: string or pathlib.Path
        Target path for a z_file output if desired
    return_collection : boolean
        return_collection=False will return an mt_metadata TF object
        return_collection=True will return
        aurora.transfer_function.transfer_function_collection.TransferFunctionCollection

    Returns
    -------
    tf: TransferFunctionCollection or mt_metadata TF
        The transfer funtion object
    tf_cls: mt_metadata.transfer_functions.TF
        TF object
    """

    # Initialize config and mth5s
    tfk = TransferFunctionKernel(dataset=tfk_dataset, config=config)
    tfk.make_processing_summary()
    tfk.validate()
    # See Note #1
    tfk.initialize_mth5s(mode="a")
    tfk.check_if_fc_levels_already_exist()  # populate the "fc" column of dataset_df
    print(f"fc_levels_already_exist = {tfk.dataset_df['fc']}")
    print(
        f"Processing config indicates {len(tfk.config.decimations)} "
        f"decimation levels "
    )

    tf_dict = {}

    for i_dec_level, dec_level_config in enumerate(tfk.valid_decimations()):
        #i_dec_level = dec_level_config.decimation.level
        tfk.update_dataset_df(i_dec_level)
        dec_level_config = tfk.apply_clock_zero(dec_level_config)

        stfts = {}
        stfts["local"] = []
        stfts["remote"] = []

        # Check first if TS processing or accessing FC Levels
        for i, row in tfk.dataset_df.iterrows():
            # This iterator could be updated to iterate over row-pairs if remote is True, corresponding to simultaneous data

            if not tfk.is_valid_dataset(row, i_dec_level):
                continue

            run_obj = row.mth5_obj.from_reference(row.run_reference)
            if row.fc:
                stft_obj = load_stft_obj_from_mth5(i_dec_level, row, run_obj)
                stfts = append_chunk_to_stfts(stfts, stft_obj, row.remote)
                continue

            run_xrds = row["run_dataarray"].to_dataset("channel")
            run_obj = row.mth5_obj.from_reference(row.run_reference)
            stft_obj = make_stft_objects(
                tfk.config, i_dec_level, run_obj, run_xrds, units, row.station_id
            )
            # Pack FCs into h5
            save_fourier_coefficients(dec_level_config, i_dec_level, row, run_obj, stft_obj)

            stfts = append_chunk_to_stfts(stfts, stft_obj, row.remote)

        local_merged_stft_obj, remote_merged_stft_obj = merge_stfts(stfts, tfk)

        # FC TF Interface here (see Note #3)
        # Could downweight bad FCs here

        ttfz_obj = process_tf_decimation_level(
            tfk.config,
            i_dec_level,
            local_merged_stft_obj,
            remote_merged_stft_obj,
        )
        ttfz_obj.apparent_resistivity(tfk.config.channel_nomenclature, units=units)
        tf_dict[i_dec_level] = ttfz_obj

        if show_plot:
            from aurora.sandbox.plot_helpers import plot_tf_obj

            plot_tf_obj(ttfz_obj, out_filename="out")

    tf_collection = TransferFunctionCollection(
        tf_dict=tf_dict, processing_config=tfk.config
    )

    tf_cls = tfk.export_tf_collection(tf_collection)

    if z_file_path:
        tf_cls.write(z_file_path)

    tfk.dataset.close_mths_objs()
    if return_collection:
        # this is now really only to be used for debugging and may be deprecated soon
        return tf_collection
    else:
        return tf_cls



"""
These methods can possibly be moved under mt_metadata, or mth5

They extract info needed to setup emtf_z files.
"""
import fortranformat as ff
import numpy as np

EMTF_CHANNEL_ORDER = ["hx", "hy", "hz", "ex", "ey"]


def get_default_orientation_block(n_ch=5):
    """
    Helper function used when working with matlab structs which do not have enough
    info to make headers

    Parameters
    ----------
    n_ch: int
        number of channels at the station

    Returns
    -------
    orientation_strs: list
        List of text strings, one per channel
    """
    orientation_strs = []
    orientation_strs.append("    1     0.00     0.00 tes  Hx\n")
    orientation_strs.append("    2    90.00     0.00 tes  Hy\n")
    if n_ch == 5:
        orientation_strs.append("    3     0.00     0.00 tes  Hz\n")
    orientation_strs.append("    4     0.00     0.00 tes  Ex\n")
    orientation_strs.append("    5    90.00     0.00 tes  Ey\n")
    return orientation_strs


def merge_tf_collection_to_match_z_file(aux_data, tf_collection):
    """
    Currently this is only used for the synthtetic test, but maybe useful for
    other tests.  Given data from a z_file, and a tf_collection,
    the tf_collection may have several TF estimates at the same frequency
    from multiple decimation levels.  This tries to make a single array as a
    function of period for all rho and phi

    Parameters
    ----------
    aux_data: merge_tf_collection_to_match_z_file
        Object representing a z-file
    tf_collection: aurora.transfer_function.transfer_function_collection
    .TransferFunctionCollection
        Object representing the transfer function returnd from the aurora processing


    Returns
    -------
    result: dict of dicts
        Keyed by ["rho", "phi"], below each of these is an ["xy", "yx",] entry.  The
        lowest level entries are numpy arrays.
    """
    rxy = np.full(len(aux_data.decimation_levels), np.nan)
    ryx = np.full(len(aux_data.decimation_levels), np.nan)
    pxy = np.full(len(aux_data.decimation_levels), np.nan)
    pyx = np.full(len(aux_data.decimation_levels), np.nan)
    dec_levels = list(set(aux_data.decimation_levels))
    dec_levels = [int(x) for x in dec_levels]
    dec_levels.sort()

    for dec_level in dec_levels:
        aurora_tf = tf_collection.tf_dict[dec_level - 1]
        indices = np.where(aux_data.decimation_levels == dec_level)[0]
        for ndx in indices:
            period = aux_data.periods[ndx]
            # find the nearest period in aurora_tf
            aurora_ndx = np.argmin(np.abs(aurora_tf.periods - period))
            rxy[ndx] = aurora_tf.rho[aurora_ndx, 0]
            ryx[ndx] = aurora_tf.rho[aurora_ndx, 1]
            pxy[ndx] = aurora_tf.phi[aurora_ndx, 0]
            pyx[ndx] = aurora_tf.phi[aurora_ndx, 1]

    result = {}
    result["rho"] = {}
    result["phi"] = {}
    result["rho"]["xy"] = rxy
    result["phi"]["xy"] = pxy
    result["rho"]["yx"] = ryx
    result["phi"]["yx"] = pyx
    return result


def clip_bands_from_z_file(z_path, n_bands_clip, output_z_path=None, n_sensors=5):
    """
    This function takes a z_file and clips periods off the end of it.
    It can come in handy sometimes -- specifically for manipulating matlab
    results of synthetic data.

    Parameters
    ----------
    z_path: Path or str
        path to the z_file to read in and clip periods from
    n_periods_clip: integer
        how many periods to clip from the end of the zfile
    overwrite: bool
        whether to overwrite the zfile or rename it
    n_sensors

    Returns
    -------

    """
    if not output_z_path:
        output_z_path = z_path

    if n_sensors == 5:
        n_lines_per_period = 13
    elif n_sensors == 4:
        n_lines_per_period = 11
        print("WARNING n_sensors==4 NOT TESTED")

    f = open(z_path, "r")
    lines = f.readlines()
    f.close()

    for i in range(n_bands_clip):
        lines = lines[:-n_lines_per_period]
    n_bands_str = lines[5].split()[-1]
    n_bands = int(n_bands_str)
    new_n_bands = n_bands - n_bands_clip
    new_n_bands_str = str(new_n_bands)
    lines[5] = lines[5].replace(n_bands_str, new_n_bands_str)

    f = open(output_z_path, "w")
    f.writelines(lines)
    f.close()
    return



"""
from egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF
/functions/Edfwts.m
"""

import numpy as np


class EffectiveDegreesOfFreedom(object):
    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs
        n_data : integer
            Number of "numeric" observations in the input and remote reference data.
            Any time-frequency index having nan value in any channel is subtracted
            from the total number of observations before computation of p1, p2
        edfl1 : int or float
             the effective number of degrees of freedom at which a datapoint should be
              removed from the total energy calculation in the iterative loop
        alpha : float
            The exponent on n_data in formulae for thresholds p1 and p2
        c1 : float
            ad hoc parameter for scaling the p1 threshold
        c2 : float
            ad hoc parameter for scaling the p2 threshold
        p3 : float
            parameter controling the threshold at which the ratio of a data weighting
            associated with remote and input channel is unaccepatble.
        """

        self.edfl1 = kwargs.get("edf_l1", 20.0)
        self.alpha = kwargs.get("alpha", 0.5)
        self.c1 = kwargs.get("c1", 2.0)
        self.c2 = kwargs.get("c2", 10.0)
        self.p3 = kwargs.get("p3", 5)
        self.n_data = kwargs.get("n_data", 0)

    @property
    def p1(self):
        """
        Threshold applied to edf.  All edf below  this value
        are set to weight=0
        """
        return self.c1 * (self.n_data ** self.alpha)

    @property
    def p2(self):
        """
        Threshold applied to edf.  All edf above th  this value
        are set to weight=0
        """
        return self.c2 * (self.n_data ** self.alpha)

    def compute_weights(self, X, use):
        """
        The data covariance matrix s and its inverse h are iteratively recomputed using
        fewer and fewer observations. However, the edf is also computed at every
        iteration but doesn't seem to use any fewer observations.
        Thus the edf weights change as use drops, EVEN FOR INDICES that were
        previously computed... this seems like it could be an error.

        Discussing this with Gary:
        "... becuase you are downweighting (omitting) more and more highpower events
        the total signal is going down.  The signal power goes down with every call
        to this method"

        Parameters
        ----------
        X
        use : boolean array

        Returns
        -------
        The "HAT" Matrix, where the diagonals of this matrix are really big it means an
        individual data point is controlling its own prediction, and the estimate.
        If the problem was balanced, each data point would contribute equally to the
        estimate.  Each data point shoudl contribute 1/N to each parameter. When one
        data point is large and the others are tiny, then it may be contributing a
        lot, say 1/2 rather than 1/n.

        edf is like the diagonal of the Hat matrix (in the single station case)
        How much does the data point contribute to the prediction of itself.
        If there are n data points contributing equally, each datapoint
        #should contribute ~1/n to its prediction

        Note: H = inv(S) in general has equal H[0,1] = H[1,0];  2x2 matrices with
        matching off-diagonal terms have inverses with the same property.

        """
        S = X[:, use] @ np.conj(X[:, use]).T  # covariance matrix, 2x2
        S /= sum(use)
        H = np.linalg.inv(S)  # invert the 2x2

        xx_term = np.real(X[0, :] * np.conj(X[0, :]) * H[0, 0])
        yy_term = np.real(X[1, :] * np.conj(X[1, :]) * H[1, 1])
        xy_term = 2 * np.real(np.conj(X[1, :]) * X[0, :] * H[1, 0])
        # ?real or abs?
        edf = xx_term + yy_term + xy_term
        return edf


def effective_degrees_of_freedom_weights(X, R, edf_obj=None, test=True):
    """
    Emulates edfwts ("effective dof") from tranmt.  Based on Edfwts.m matlab code from
    iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/functions/

    Flow:
    0. Initialize weights vector (of 1's) the length of the "observation" axis
    1. Remove any nan in X,R
    2. compute the weights on the reduced (no-nan) arrays of X and R
    3. Overwrite the weights vector for the non-nan entries.
    4 Return the weights vector.  We then broadcast multiply that array against the
    data, to apply the weights.

    This follows the matlab code by using a boolean index vector.  An xarray
    implementaiton which uses the observation dimension of X, R was started but never
    finished.  Here are the breadcrumbs for the xarray method:
    0. Create weights:
    import xarray as xr
    X = X.assign(weights=lambda x: X.frequency*0+1.0)
    weights = X["weights"]
    X = X[["hx","hy"]]
    1. Then to drop nan:
    from aurora.time_series.xarray_helpers import handle_nan
    X, Y, RR = handle_nan(XX, None, R, drop_dim="observation")
    2. Same as below
    3. Overwrite: Not sure, it wasn't clear how to assign them. Possibly the syntax
    with multi-index was confusing me when trying to assign.
    Accessing the weights was done by:
    weights.sel(observation=X.observation)
    But the assignment syntax I'm not sure about.  Could also be that I need to cast
    as dataarray for the assignement step.

    Things to review:
    -Why is the Remote reference weighting not done with a while loop?  This maybe
    just an oversight in the matlab codes.
    - since we assign zero-weights to Nan, we could probably remove the keep_indices
    methods and simply assign nan or zero weights to those data up front. Since the
    "use" boolean selects data before computiation are performed on X, R we should
    never get nans in the computed edfs, but should simplify the code somewhat.

    Parameters
    ----------
    X : xr.Dataset
        The input channels for regression.  Usually horizontal magnetics hx, hy
    R : xr.Dataset or None
        The remote reference channels, usually hx, hy.  Can be None if single-station
        processing
    EDFparam : EffectiveDegreesOfFreedom
        Object with parameters for Gary's adhoc edfwts method.

    Returns
     weights : numpy array
        Weights for reducing leverage points.
    -------

    """
    num_channels = len(X.data_vars)
    if num_channels != 2:
        print("edfwts only works for 2 input channels")
        raise Exception
    X = X.to_array(dim="channel")
    if R is not None:
        R = R.to_array(dim="channel")

    if test:
        pass
        # add_nan - this syntax only works for dataarray
        # X[0,10] *= 2e1
        # X[0,1] *= np.nan
        # # X[0,-3] *= np.nan
        # # try:
        # #     X[0,-13] *= 1e13#np.nan
        # # except IndexError:
        # #     pass
        # # if R is not None:
        # #     R[1,2:11] *= np.nan

    n_observations_initial = len(X.observation)
    weights = np.ones(n_observations_initial)

    # reduce the data to only valid (non-nan) observations
    if R is not None:
        keep_x_indices = ~np.isnan(X.data).any(axis=0)
        keep_r_indices = ~np.isnan(R.data).any(axis=0)
        keep_indices = keep_r_indices & keep_x_indices
        RR = R.data[:, keep_indices]
    else:
        keep_indices = ~np.isnan(X.data).any(axis=0)
    n_observations_numeric = keep_indices.sum()
    XX = X.data[:, keep_indices]

    if edf_obj is None:
        edf_obj = EffectiveDegreesOfFreedom(n_data=n_observations_numeric)

    # %    determine intial robust B-field cross-power matrix; this just uses
    # %    edfl1 -- cut off for estimating robust local magnetic covariance
    nOmit = n_observations_numeric
    use = np.ones(n_observations_numeric, dtype=bool)
    # initialize use as a boolean of True, as large as XX
    n_valid_observations = n_observations_numeric
    # "use" never changes length, it only flips its bit
    # The while loop exits when n_valid_observations == sum(use)
    # i.e.the effective dof are all below threshold
    # Estimate dof.  Then we "use" only points whose dof are smaller than the
    # threshold.  Then we recompute dof.  This time the covariance matrix diagonals
    # are smaller, there is less energy in the time series for the S, H calculation.
    while nOmit > 0:
        eff_deg_of_freedom = edf_obj.compute_weights(XX, use)
        use = eff_deg_of_freedom <= edf_obj.edfl1  # update "use" boolean selector
        nOmit = n_valid_observations - sum(use)
        n_valid_observations = sum(use)

    wt = np.ones(n_observations_numeric)
    wt[eff_deg_of_freedom > edf_obj.p2] = 0
    cond = (eff_deg_of_freedom <= edf_obj.p2) & (eff_deg_of_freedom > edf_obj.p1)
    wt[cond] = np.sqrt(edf_obj.p1 / eff_deg_of_freedom[cond])

    # May want to add a loop in the RR
    if R is not None:
        # now find additional segments with crazy remotes
        wtRef = np.ones(n_observations_numeric)
        ref_use = np.ones(n_observations_numeric, dtype=bool)
        edf_ref = edf_obj.compute_weights(RR, ref_use)

        wtRef[edf_ref > edf_obj.p2] = 0
        cond = (edf_ref <= edf_obj.p2) & (edf_ref > edf_obj.p1)
        wtRef[cond] = np.sqrt(edf_obj.p1 / edf_ref[cond])

        # are either of the weights very different
        cond1 = wtRef / wt > edf_obj.p3
        cond2 = wt / wtRef > edf_obj.p3
        differentAmp = cond1 | cond2
        wt = wt * wtRef
        wt[differentAmp] = 0

    weights[keep_indices] = wt
    weights[weights == 0] *= np.nan
    return weights



"""
follows Gary's TRME.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes

Note that the matlab code claimed:
    %  R2 is squared coherence (top row is using raw data, bottom
    %    cleaned, with crude correction for amount of downweighted data)
However, that is not the case. There was only one estimate for R2 in the matlab codes

"""
import numpy as np
import xarray as xr

from aurora.transfer_function.regression.m_estimator import MEstimator


class TRME_RR(MEstimator):
    def __init__(self, **kwargs):
        """
        Robust remote reference estimator.  Z is the reference station data,
        and is the same size as X (see regression.base.RegressionEstimator(
        """
        super(TRME_RR, self).__init__(**kwargs)
        self._Z = kwargs.get("Z", None)
        self.Z = self._Z.to_array().data.T
        self.qr_input = "Z"
        self._QHX = None
        self.check_for_nan()
        self.check_number_of_observations_xy_consistent()
        self.check_for_enough_data_for_rr_estimate()
        self.check_reference_data_shape()

    def check_for_nan(self):
        cond1 = np.isnan(self.X).any()
        cond2 = np.isnan(self.Y).any()
        cond3 = np.isnan(self.Z).any()
        nans_present = cond1 or cond2 or cond3
        if nans_present:
            print("Missing data not allowed for TRME_RR class")
            raise Exception

    def check_for_enough_data_for_rr_estimate(self):
        if self.is_underdetermined:
            error_msg = "not enough data for RR estimate:"
            error_msg = f"{error_msg} n_channels_in = {self.n_channels_in}"
            error_msg = f"{error_msg} N_data = {self.n_data}"
            print(f"{error_msg}")
            raise Exception

    def check_reference_data_shape(self):
        if self.Z.shape != self.X.shape:
            print("sizes of local and remote do not agree in RR estimation routine")
            raise Exception

    def update_y_hat(self):
        self._Y_hat = self.X @ self.b
        return

    def update_b(self):
        """matlab was: b = QTX\QTY"""
        self.b = np.linalg.solve(self.QHX, self.QHYc)
        return

    @property
    def QHX(self):
        if self._QHX is None:
            self._QHX = self.QH @ self.X
        return self._QHX

    def update_residual_variance(self, correction_factor=1):
        self._residual_variance = self.residual_variance_method1()
        self._residual_variance *= correction_factor
        return self._residual_variance

    def compute_inverse_signal_covariance(self):
        """
        Matlab code was :
        Cov_SS = (self.ZH @self.X) \ (self.XH @ self.X) / (self.XH @self.Z)
        I broke the above line into B/A where
        B = (self.ZH @self.X) \ (self.XH @ self.X), and A = (self.XH @self.Z)
        Then follow matlab cookbook, B/A for matrices = (A'\B')
        """
        ZH = self.Z.conj().T
        XH = self.X.conj().T
        B = np.linalg.solve(ZH @ self.X, XH @ self.X)
        A = XH @ self.Z
        cov_ss_inv = np.linalg.solve(A.conj().T, B.conj().T).conj().T

        self.cov_ss_inv = xr.DataArray(
            cov_ss_inv,
            dims=["input_channel_1", "input_channel_2"],
            coords={
                "input_channel_1": list(self._X.data_vars),
                "input_channel_2": list(self._X.data_vars),
            },
        )
        return



import numpy as np
import scipy.signal as ssig

from scipy.interpolate import interp1d

from aurora.time_series.decorators import can_use_xr_dataarray


def validate_coordinate_ordering_time_domain(dataset):
    """
    Check that the data dimensions are what you expect.  This may evolve, but for now,
    just want to make sure that we are operating along the correct axes when we
    demean, detrend, taper, etc.
    Parameters
    ----------
    dataset : xarray.Dataset

    Returns
    -------

    """
    coordinate_labels = list(dataset.coords.keys())
    cond1 = coordinate_labels[0] == "within-window time"
    cond2 = coordinate_labels[1] == "time"
    if cond1 & cond2:
        return True
    else:
        print("Uncertain that xarray coordinates are correctly ordered")
        raise Exception


def get_time_coordinate_axis(dataset):
    """
    It is common to pass an argument to scipy.signal methods axis=int
    where that integer specifies along which dimension we are applying the
    operator.  This method helps ensure that we have the correct axis.
    Parameters
    ----------
    dataset : xarray.Dataset

    Returns
    -------

    """
    coordinate_labels = list(dataset.coords.keys())

    if len(coordinate_labels) != 2:
        print("Warning - Expected two distinct coordinates")
        # raise Exception

    return coordinate_labels.index("time")
    # time_coord_indices = [ndx for x, ndx in enumerate(coordinate_labels) if
    #                                                   x=="time"]
    # if len(time_coord_indices) == 1:
    #     return time_coord_indices[0]
    # else:
    #     print("expected only one universal time coordinate")
    #     raise Exception


class WindowedTimeSeries(object):
    """
    Time series that has been chopped into (possibly) overlapping windows.

    This is a place where we can put methods that operate on these sorts of
    objects.

    The assumption is that we take xarrays keyed by "channel"

    Specific methods:
        Demean
        Detrend
        Prewhiten
        stft
        invert_prewhitening

        probably make these @staticmethod s so we import WindowedTimeSeries
        and then call the static methods
    """

    def __init__(self):
        pass

    @can_use_xr_dataarray
    @staticmethod
    def apply_taper(data=None, taper=None, in_place=True):
        """
        Point by point multiplication of taper against time series.
        xarray handles this very cleanly as a direct multiply operation.
        tapered_obj = windowed_obj * windowing_scheme.taper
        """
        data = data * taper
        return data

    @staticmethod
    def detrend(data=None, detrend_axis=None, detrend_type=None, inplace=True):
        """
        Notes: overwrite data=True probably best for most applications but be careful
         with that.  Do we want to avoid this in general? Could we be possibly
         overwriting stuff on MTH5 in future?
         Also, is overwrite even working how I think it is here?
         Overwrite_data not working right in scipy.signal, dont use it for now

        Parameters
        ----------
        data : xarray Dataset
        detrend_axis : string
        detrend_type : string
            "linear" or "constant"
            This argument is provided to scipy.signal.detrend

        Returns
        -------

        """
        if detrend_axis is None:
            detrend_axis = get_time_coordinate_axis(data)
        if not inplace:
            raise NotImplementedError

        for channel in data.keys():

            # windowed_array = data[key].data
            nanless_data = data[channel].dropna(dim="time")
            ensembles = nanless_data.data
            if detrend_type:  # neither False nor None
                try:
                    ensembles = ssig.detrend(
                        ensembles, axis=detrend_axis, type=detrend_type
                    )
                    # overwrite_data=True
                except ValueError as error:
                    msg = (
                        "Could not detrend "
                        f"{channel} in time range "
                        f"{data[channel].coords.indexes['time'][0].isoformat()} to "
                        f"{data[channel].coords.indexes['time'][-1].isoformat()}."
                    )
                    if ensembles.size == 0:
                        print(msg + " NO DATA")
                    else:
                        print(msg + "UNKOWN REASON:" + error)

            if inplace:
                if len(nanless_data.time) < len(data[channel].time):
                    data[channel].data += np.nan
                    data[channel].loc[nanless_data.time, :] = ensembles
                    # there must be a cute way to just assign all nan to just the
                    # columns that had nan values in some places, rather than
                    # assigning nan to the whole array and then overwriting the
                    # ensembles .. something like
                    # data[channel].loc[~nanless_data.time, :] = np.nan
                else:
                    data[channel].data = ensembles
            else:
                raise NotImplementedError
        return data

    @staticmethod
    def apply_stft(
        data=None,
        sample_rate=None,
        detrend_type=None,
        spectral_density_calibration=1.0,
        fft_axis=None,
    ):
        """
        Only supports xr.Dataset at this point

        Parameters
        ----------
        data
        sample_rate
        detrend_type

        Returns
        -------

        """
        from aurora.time_series.windowing_scheme import fft_xr_ds

        if fft_axis is None:
            fft_axis = get_time_coordinate_axis(data)
        spectral_ds = fft_xr_ds(data, sample_rate, detrend_type=detrend_type)
        spectral_ds *= spectral_density_calibration
        return spectral_ds

    def delay_correction(self, dataset, run_obj):
        """

        Parameters
        ----------
        dataset : xr.Dataset
        run_obj :

        Returns
        -------

        """
        # "NOT TESTED - PSEUDOCODE ONLY"
        for channel_id in dataset.keys():
            mth5_channel = run_obj.get_channel(channel_id)
            channel_filter = mth5_channel.channel_response_filter
            delay_in_seconds = channel_filter.total_delay
            true_time_axis = dataset.time + delay_in_seconds
            interpolator = interp1d(
                true_time_axis, dataset[channel_id].data, assume_sorted=True
            )
            corrected_data = interpolator(dataset.time)
            dataset[channel_id].data = corrected_data
        return dataset



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 17 14:15:20 2022

@author: jpeacock
"""
# =============================================================================
# Imports
# =============================================================================
import pandas as pd

from mt_metadata.base.helpers import write_lines
from mt_metadata.base import get_schema, Base
from mt_metadata.timeseries import TimePeriod
from .standards import SCHEMA_FN_PATHS
from .run import Run

# =============================================================================
attr_dict = get_schema("station", SCHEMA_FN_PATHS)
# =============================================================================


class Station(Base):
    __doc__ = write_lines(attr_dict)

    def __init__(self, **kwargs):
        super().__init__(attr_dict=attr_dict, **kwargs)
        self._runs = []

    @property
    def runs(self):
        return self._runs

    @runs.setter
    def runs(self, values):
        self._runs = []
        if not isinstance(values, list):
            values = [values]

        for item in values:
            if isinstance(item, str):
                run = Run(id=item)
            elif isinstance(item, Run):
                run = item
            elif isinstance(item, dict):
                run = Run()
                run.from_dict(item)

            else:
                raise TypeError(f"not sure what to do with type {type(item)}")

            self._runs.append(run)

    def get_run(self, run_id):
        """

        :param run_id: DESCRIPTION
        :type run_id: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        try:
            return self.run_dict[run_id]
        except KeyError:
            raise KeyError(f"Could not find {run_id}")

    @property
    def run_list(self):
        """list of run names"""

        return [r.id for r in self.runs]

    @property
    def run_dict(self):
        """
        need to have a dictionary, but it can't be an attribute cause that
        gets confusing when reading in a json file

        :return: DESCRIPTION
        :rtype: TYPE

        """
        return dict([(rr.id, rr) for rr in self.runs])

    def to_dataset_dataframe(self):
        """
        Create a dataset definition dataframe that can be used in the
        processing

        [
            "station_id",
            "run_id",
            "start",
            "end",
            "mth5_path",
            "sample_rate",
            "input_channels",
            "output_channels",
            "remote",
        ]

        """

        data_list = []

        for run in self.runs:
            for tp in run.time_periods:
                entry = {
                    "station_id": self.id,
                    "run_id": run.id,
                    "start": tp.start,
                    "end": tp.end,
                    "mth5_path": self.mth5_path,
                    "sample_rate": run.sample_rate,
                    "input_channels": run.input_channel_names,
                    "output_channels": run.output_channel_names,
                    "remote": self.remote,
                    "channel_scale_factors": run.channel_scale_factors,
                }
                data_list.append(entry)

        df = pd.DataFrame(data_list)
        df.start = pd.to_datetime(df.start)
        df.end = pd.to_datetime(df.end)

        return df

    def from_dataset_dataframe(self, df):
        """
        set a data frame

        [
            "station_id",
            "run_id",
            "start",
            "end",
            "mth5_path",
            "sample_rate",
            "input_channels",
            "output_channels",
            "remote",
        ]

        :param df: DESCRIPTION
        :type df: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        self.runs = []

        self.id = df.station_id.unique()[0]
        self.mth5_path = df.mth5_path.unique()[0]
        self.remote = df.remote.unique()[0]

        for entry in df.itertuples():
            try:
                r = self.run_dict[entry.run_id]
                r.time_periods.append(
                    TimePeriod(start=entry.start.isoformat(), end=entry.end.isoformat())
                )

            except KeyError:
                if hasattr(entry, "channel_scale_factors"):
                    channel_scale_factors = entry.channel_scale_factors
                else:
                    channel_scale_factors = {}
                r = Run(
                    id=entry.run_id,
                    sample_rate=entry.sample_rate,
                    input_channels=entry.input_channels,
                    output_channels=entry.output_channels,
                    channel_scale_factors=channel_scale_factors,
                )

                r.time_periods.append(
                    TimePeriod(start=entry.start.isoformat(), end=entry.end.isoformat())
                )
                self.runs.append(r)



"""
based on Gary's
RhoPlot.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes
"""
import matplotlib.pyplot as plt
import numpy as np

from aurora.transfer_function.plot.error_bar_helpers import err_log

plt.ioff()


class RhoPlot(object):
    """
    TF plotting object class; some methods are only relevant to
    specific types of TFs (or for derived parameters such as rho/phi)

    """

    def __init__(self, tf_obj):
        self.tf = tf_obj

    def phase_plot(self, pred=None):
        """
        simple driver to plot phase only on a separate figures of fixed size

        hfig = phasePlot(obj, pred)
        hfig = figure('Position', [100, 100, 325, 200],
        'PaperPosition',[1, 1, 3.25, 2.], 'Tag', 'Phase Plot');
        axRect = [0.1446    0.2150    0.7604    0.7100];
        if nargin == 2
            phaseSubPlot(obj, hfig, axRect, pred);
        else
            phaseSubPlot(obj, hfig, axRect);
        end

        Returns
        -------

        """
        axRect = [0.1446, 0.2150, 0.7604, 0.7100]
        # plt.figure(22, figsize = (8.5, 11), dpi=300)
        fig, ax = plt.subplots()
        if pred is not None:
            self.phase_sub_plot(ax, axRect, pred=pred)
        else:
            self.phase_sub_plot(ax, axRect)

    def rho_plot2(self, pred=None):
        """
        one off, use tools in MTpy

        """
        axRect = [0.1446, 0.2150, 0.7604, 0.7100]
        # plt.figure(22, figsize = (8.5, 11), dpi=300)
        fig, ax = plt.subplots()
        if pred is not None:
            self.rho_sub_plot(ax, axRect, pred=pred)
        else:
            self.rho_sub_plot(ax, axRect)

    def rho_phi_plot(self, pred=None):
        """

        Parameters
        ----------
        pred

        Returns
        -------

        """
        rects = self.set_figure_size()
        fig, ax = plt.subplots(2, 1, 1)
        # hfig = figure('Position', rects.Screen, 'PaperPosition',
        #                rects.Paper, 'Tag', 'rho-phi plot');

        if pred is not None:
            h_phi = self.phase_sub_plot(ax, rects.phi, pred)
            h_rho = self.rho_sub_plot(ax, rects.rho, pred)
        else:
            h_phi = self.phase_sub_plot(ax, rects.phi)
            h_rho = self.rho_sub_plot(ax, rects.rho)

        H = {"fig": fig, "rho": h_rho, "phi": h_phi}
        return H

    # def rho_plot(self, **kwargs):
    #     """
    #     simple driver to plot rho only on a separate figures of fixed size
    #     function[hfig, hax] = rhoPlot(obj, varargin)
    #     Returns
    #     -------
    #
    #     """
    #     AddPred = 0
    #     inds = 1
    #     # <HANDLE KWARGS>
    #     # n_kwargs = len(**kwargs)
    #     # if np.mod(n_kwargs, 2)==1:
    #     #     print('Optional arguments must occur in pairs')
    #     #     raise Exception
    #     #
    #     # n_kwarg_pairs = n_kwargs/2
    #     # for k in range(0, n_kwargs, 2):
    #     #     option = lower(varargin{k});
    #     #     switch option
    #     #     case 'pred'
    #     #     AddPred = varargin{k + 1};
    #     #     case 'inds'
    #     #     inds = varargin{k + 1};
    #     # <HANDLE KWARGS>
    #     rects = self.set_figure_size()
    #     yFac = 1 - (rects.phi(1) + rects.phi(3))
    #     yMarg = (rects.rho(1) - (rects.phi(1) + rects.phi(3))) / yFac
    #     rectScreen = rects.Screen
    #     rectScreen[3] = rectScreen[3] * yFac
    #     rectPaper = rects.Paper
    #     rectPaper[3] = rectPaper[3] * yFac
    #     rectRho = rects.Rho
    #     rectRho[3] = rectRho[3] / yFac
    #     rectRho[1] = yMarg
    #     hfig = plt.figure(
    #         "Position", rectScreen, "PaperPosition", rectPaper, "Tag", "rho Plot"
    #     )
    #     if AddPred:
    #         hax = self.rho_sub_plot(hfig, rectRho, pred)
    #     else:
    #         if inds == 1:
    #             hax = self.rho_sub_plot(hfig, rectRho)
    #         else:
    #             hax = self.rho_sub_plots(hfig, inds)
    #     # grid on

    def phase_sub_plot(self, ax, ttl_str="", axRect=None, pred=None):
        """
        place a phase subplot on a figure, given figure handle and axis
        postion

        Parameters
        ----------
        self
        ax
        axRect
        pred

        Returns
        -------

        """

        phi = self.tf.phi
        # rotate phases so all are positive:
        negative_phi_indices = np.where(phi < 0)[0]
        phi[negative_phi_indices] += 180.0

        Tmin, Tmax = self.set_period_limits()
        axis_limits = [Tmin, Tmax, 0, 90]
        # xticks = self.get_xticks()

        [xb, yb] = err_log(
            np.transpose(self.tf.periods),
            self.tf.phi[:, 0],
            self.tf.phi_se[:, 0],
            "XLOG",
            axis_limits,
        )
        # figure(ax); #need this?
        # set current axes
        # hax = plt.axes('Position', axRect);
        ax.semilogx(xb, yb, "b-")
        ax.semilogx(self.tf.periods, phi[:, 0], "bo")
        # print("OK, now set linewidth and markersize")
        # set(lines, 'LineWidth', 1, 'MarkerSize', 7);
        # hold on;
        xb, yb = err_log(
            np.transpose(self.tf.periods),
            self.tf.phi[:, 1],
            self.tf.phi_se[:, 1],
            "XLOG",
            axis_limits,
        )
        ax.semilogx(xb, yb, "r-")
        ax.semilogx(self.tf.periods, phi[:, 1], "ro")
        # set(lines, 'LineWidth', 1, 'MarkerSize', 7);
        if pred is not None:
            plt.plot(pred.tf.periods, pred.tf.phi[:, 0], "b-", "linewidth", 2)
            plt.plot(pred.tf.periods, pred.tf.phi[:, 1], "r-", "linewidth", 2)

        # (lims_ph);
        ax.set_xlim(axis_limits[0], axis_limits[1])
        ax.set_ylim(axis_limits[2], axis_limits[3])
        title_pos_x = np.log(axis_limits[0]) + 0.1 * (
            np.log(axis_limits[1] / axis_limits[0])
        )
        title_pos_x = np.ceil(np.exp(title_pos_x))
        title_pos_y = axis_limits[2] + 0.8 * (axis_limits[3] - axis_limits[2])
        # ttl_str = f"$\phi$ : {self.tf.header.local_station_id}"\
        # + \"PKD"#self.tf.Header.LocalSite.SiteID
        ax.text(title_pos_x, title_pos_y, ttl_str, fontsize=14, fontweight="demi")
        # set(gca, 'FontWeight', 'bold', 'FontSize', 11, 'Xtick', xticks);
        ax.set_xlabel("Period (s)")
        ax.set_ylabel("Degrees")

    def rho_sub_plot(self, ax, ttl_str="", axRect=None, pred=None):
        """
        Calls plotrhom, standard plotting routine; uses some other routines in
        EMTF/matlab/Zplt; this version is for putting multiple curves on the
        same plot

        set plotting limits now that rho is known
        Parameters
        ----------
        ax
        axRect
        pred

        Returns
        -------

        """
        lims, orient = self.set_lims()
        # lims_rho = lims[0:3];
        axis_limits = lims[0:4]
        # xticks = self.get_xticks()
        # plot error bars:
        [xb, yb] = err_log(
            self.tf.periods,
            self.tf.rho[:, 0],
            self.tf.rho_se[:, 0],
            "XLOG",
            axis_limits,
        )
        ax.loglog(xb, yb, "b-")

        # plot rho dots
        ax.loglog(self.tf.periods, self.tf.rho[:, 0], "bo")

        [xb, yb] = err_log(
            self.tf.periods,
            self.tf.rho[:, 1],
            self.tf.rho_se[:, 1],
            "XLOG",
            axis_limits,
        )
        ax.loglog(xb, yb, "r-")
        ax.loglog(self.tf.periods, self.tf.rho[:, 1], "ro")

        if pred is not None:
            plt.plot(pred.tf.periods, pred.tf.rho[:, 0], "b-", "linewidth", 1.5)
            plt.plot(pred.tf.periods, pred.tf.rho[:, 1], "r-", "linewidth", 1.5)

        # axis(lims_rho);
        ax.set_xlim(axis_limits[0], axis_limits[1])
        ax.set_ylim(axis_limits[2], axis_limits[3])

        #
        title_pos_x = np.log(axis_limits[0]) + 0.1 * (
            np.log(axis_limits[1] / axis_limits[0])
        )
        title_pos_x = np.ceil(np.exp(title_pos_x))
        title_pos_y = axis_limits[2] + 0.8 * (axis_limits[3] - axis_limits[2])
        ttl_str = "\u03C1_a : " + ttl_str
        # c_title = "$\rho_a$ :" + "PKD"  # obj.tf.Header.LocalSite.SiteID
        ax.text(title_pos_x, title_pos_y, ttl_str, fontsize=14, fontweight="demi")
        # set(gca, 'FontWeight', 'bold', 'FontSize', 11, 'Xtick', xticks);
        ax.set_xlabel("Period (s)")
        ax.set_ylabel("$\Omega$-m")
        return

    def set_period_limits(self):  # function[Tmin, Tmax] = setTlim(obj)
        """
        set nicer period limits for logartihmic period scale plots
        Returns
        -------

        """

        x_min = self.tf.minimum_period
        x_max = self.tf.maximum_period

        Tmin = 10 ** (np.floor(np.log10(x_min) * 2) / 2)
        if (np.log10(x_min) - np.log10(Tmin)) < 0.15:
            Tmin = 10 ** (np.log10(Tmin) - 0.3)

        Tmax = 10 ** (np.ceil(np.log10(x_max) * 2) / 2)
        if (np.log10(Tmax) - np.log10(x_max)) < 0.15:
            Tmax = 10 ** (np.log10(Tmax) + 0.3)
        return Tmin, Tmax

    def set_lims(self):
        """
        set default limits for plotting; QD, derived from ZPLT
         use max/min limits of periods, rho to set limits

        function[lims, orient] = set_lims(obj)
        Returns
            lims : list
            x_max, x_min, y_max, y_min, 0, 90

            orient: 0

        TODO: maybe set this as a class?
        -------

        """
        xx_min, xx_max = self.set_period_limits()
        y_min = self.tf.rho.min()
        y_min = max(y_min, 1e-20)
        y_max = self.tf.rho.max()
        y_max = max(y_max, 1e-20)

        yy_min = 10 ** (np.floor(np.log10(y_min)))
        if (np.log10(y_min) - np.log10(yy_min)) < 0.15:
            yy_min = 10 ** (np.log10(yy_min) - 0.3)

        yy_max = 10 ** (np.ceil(np.log10(y_max)))
        if (np.log10(yy_max) - np.log10(y_max)) < 0.15:
            yy_max = 10 ** (np.log10(yy_max) + 0.3)

        if abs(yy_max - yy_min) > 1:
            lims = [xx_min, xx_max, yy_min, yy_max, 0, 90]
        else:
            lims = [xx_min, xx_max, 0.01, 1e4, 0, 90]

        orient = 0.0
        return lims, orient

    def set_figure_size(self):
        """
        rects is a dict
        Returns:
            rects : dict
            has keys: "Screen", "Paper", "Rho", "Phi"
        -------

        """
        lims, _ = self.set_lims()
        size_fac = 50
        paperSizeFac = 0.65
        one_dec = 1.6
        xdecs = np.log10(lims(1)) - np.log10(lims(0))
        one_dec = one_dec * 4 / xdecs
        ydecs = np.log10(lims[3]) - np.log10(lims[2])
        paper_width = xdecs * one_dec
        paper_height = (ydecs + 3) * one_dec
        paper_height = min([paper_height, 9])
        rectScreen = [0.5, 0.5, paper_width, paper_height] * size_fac
        rectPaper = [1.0, 1.0, paper_width * paperSizeFac, paper_height * paperSizeFac]

        rectRho = [0.15, 0.15 + 2.3 / (ydecs + 3), 0.8, ydecs / (ydecs + 3) * 0.8]
        rectPhi = [0.15, 0.15, 0.8, 2 / (ydecs + 3) * 0.8]
        rects = {
            "Screen": rectScreen,
            "Paper": rectPaper,
            "Rho": rectRho,
            "Phi": rectPhi,
        }
        return rects

    def get_xticks(self):
        xticks = 10.0 ** np.arange(-5, 6)
        cond1 = xticks >= self.tf.minimum_period
        cond2 = xticks <= self.tf.maximum_period
        xticks = xticks[cond1 & cond2]
        return xticks


def test():
    pass


def main():
    test()


if __name__ == "__main__":
    main()



import logging
import pathlib
import unittest

from aurora.general_helper_functions import count_lines
from aurora.general_helper_functions import DotDict
from aurora.general_helper_functions import TEST_PATH


class TestGeneralHelperFunctions(unittest.TestCase):
    """ """

    def setUp(self):
        logging.getLogger("matplotlib.font_manager").disabled = True
        logging.getLogger("matplotlib.ticker").disabled = True

    def test_count_lines(self):
        tmp_file = TEST_PATH.joinpath("tmp.txt")
        n_lines_in = 42
        lines = n_lines_in * ["test\n"]
        f = open(tmp_file, "w")
        f.writelines(lines)
        f.close()
        n_lines_out = count_lines(tmp_file)
        assert n_lines_out == n_lines_in
        tmp_file.unlink()
        return

    def test_dot_dict(self):
        tmp = {}
        tmp["a"] = "aa"
        tmp["b"] = "bb"
        dot_dict = DotDict(tmp)
        assert dot_dict.a == tmp["a"]
        assert dot_dict.b == "bb"


def main():
    # tmp = TestMetadataValuesSetCorrect()
    # tmp.setUp()
    # tmp.test_start_times_correct()
    unittest.main()


if __name__ == "__main__":
    main()



"""
@author: kkappler

Module to manage windowing prior to FFT.  Intended to support most
apodization  windows available via scipy.signal.get_window()


    Supported Window types = ['boxcar', 'triang', 'blackman', 'hamming', 'hann',
      'bartlett', 'flattop', 'parzen', 'bohman', 'blackmanharris',
      'nuttall', 'barthann', 'kaiser', 'gaussian', 'general_gaussian',
      'slepian', 'chebwin']

    have_additional_args = {
      'kaiser' : 'beta',
      'gaussian' : 'std',
      'general_gaussian' : ('power', 'width'),
      'slepian' : 'width',
      'chebwin' : 'attenuation',
    }

The Taper Config has 2 possible forms:
1. Standard form for accessing scipy.signal:
["taper_family", "num_samples_window", "additional_args"]
2. User-defined : for defining custom tapers

Example 1 : Standard form
"taper_family" = "hamming"
"num_samples_window" = 128
"additional_args" = {}

Example 2 : Standard form
"taper_family" = "kaiser"
"num_samples_window" = 64
"additional_args" = {"beta":8}

Examples 3 : User Defined
2. user-defined: ["array"]
In this case num_samples_window is defined by the array.
"array" = [1, 2, 3, 4, 5, 4, 3, 2, 1]
If "array" is non-empty then assume the user-defined case.

It is a little bit unsatisfying that the args need to be ordered for
scipy.signal.get_window().  Probably use OrderedDict()
for any windows that have more than one additional args.

For example
"taper_family" = 'general_gaussian'
"additional_args" = OrderedDict("power":1.5, "sigma":7)

"""


import numpy as np
import scipy.signal as ssig


class ApodizationWindow(object):
    """
    Instantiate an apodization window object.  Example usages:
    apod_window = ApodizationWindow()
    taper=ApodizationWindow(taper_family='hanning', num_samples_window=55 )

    Window factors S1, S2, CG, ENBW are modelled after Heinzel et al. p12-14
    [1] Spectrum and spectral density estimation by the Discrete Fourier transform
    (DFT), including a comprehensive list of window functions and some new
    flat-top windows.  G. Heinzel, A. Roudiger and R. Schilling, Max-Planck
    Institut fur Gravitationsphysik (Albert-Einstein-Institut)
    Teilinstitut Hannover February 15, 2002
    See Also
    [2] Harris FJ. On the use of windows for harmonic analysis with the discrete
    Fourier transform. Proceedings of the IEEE. 1978 Jan;66(1):51-83.

    Nomenclature from Heinzel et al.
    ENBW: Effective Noise BandWidth, see Equation (22)
    NENBW Normalized Equivalent Noise BandWidth, see Equation (21)

    Parameters
    ----------
    taper_family : string
        Specify the taper type - boxcar, kaiser, hanning, etc
    num_samples_window : int
        The number of samples in the taper
    taper : numpy array
        The actual window coefficients themselves.  This can be passed if a
        particular custom window is desired.
    additional_args: dictionary
        These are any additional requirements scipy needs in order to
        generate the window.
    """

    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs: dict
            See parameters list in class level documentation above
        """
        self.taper_family = kwargs.get("taper_family", "boxcar")
        self._num_samples_window = kwargs.get("num_samples_window", 0)
        self._taper = kwargs.get("taper", np.empty(0))
        self.additional_args = kwargs.get("taper_additional_args", {})

        self._coherent_gain = None
        self._nenbw = None
        self._S1 = None
        self._S2 = None
        self._apodization_factor = None

        if self.taper.size == 0:
            self.make()

    @property
    def summary(self):
        """

        Returns
        -------
        out_str: str
            String comprised of the taper_family, number_of_samples, and True/False
            if self.taper is not None
        """
        self.test_linear_spectral_density_factor()
        string1 = f"{self.taper_family} {self.num_samples_window}"
        string1 += f" taper_exists = {bool(self.taper.any())}"
        string2 = f"NENBW = {self.nenbw:.3f}, CG = {self.coherent_gain:.3f},  "
        string2 += f"window factor = {self.apodization_factor:.3f}"
        out_str = "\n".join([string1, string2])
        return out_str

    def __str__(self):
        """
        ? __repr__?
        """
        return f"{self.taper}"

    @property
    def num_samples_window(self):
        if self._num_samples_window == 0:
            self._num_samples_window = len(self.taper)
        return self._num_samples_window

    def make(self):
        """
        this is just a wrapper call to scipy.signal
        Note: see scipy.signal.get_window for a description of what is
        expected in args[1:]. http://docs.scipy.org/doc/scipy/reference/
        generated/scipy.signal.get_window.html

        note: this is just repackaging the args so that scipy.signal.get_window()
        accepts all cases.
        """
        window_args = [v for k, v in self.additional_args.items()]
        window_args.insert(0, self.taper_family)
        window_args = tuple(window_args)
        # print(f"\n\nWINDOW args {window_args}")
        self.taper = ssig.get_window(window_args, self.num_samples_window)
        self.apodization_factor  # calculate
        return

    @property
    def S1(self):
        """sum of the window coefficients"""
        if getattr(self, "_S1", None) is None:
            self._S1 = sum(self.taper)
        return self._S1

    @property
    def S2(self):
        """sum of squares of the window coefficients"""
        if getattr(self, "_S2", None) is None:
            self._S2 = sum(self.taper**2)
        return self._S2

    @property
    def coherent_gain(self):
        """DC gain of the window normalized by window length"""
        return self.S1 / self.num_samples_window

    @property
    def nenbw(self):
        """NENBW Normalized Equivalent Noise BandWidth, see Equation (21) in
        Heinzel et al 2002"""
        return self.num_samples_window * self.S2 / (self.S1**2)

    def enbw(self, fs):
        """
        Notes that unlike NENBW, CG, S1, S2, this is not a pure property of the
        window -- but instead this is a property of the window combined with
        the sample rate.
        Parameters
        ----------
        fs : sampling frequency (1/dt)

        Returns
        -------

        """
        """Effective Noise BandWidth = fs*NENBW/N = fs S2/(S1**2)"""

        return fs * self.S2 / (self.S1**2)

    def test_linear_spectral_density_factor(self):
        """This is just a test to verify some algebra
        Claim:
        The lsd_calibration factors
        A      (1./coherent\_gain)\*np.sqrt((2\*dt)/(nenbw\*N))
        and
        B      np.sqrt(2/(sample\_rate\*self.S2))
        are identical.

        Note sqrt(2\*dt)==sqrt(2\*sample_rate) so we can cancel these terms and
        A=B IFF
        (1./coherent\_gain) * np.sqrt(1/(nenbw\*N)) == 1/np.sqrt(S2)
        which I show in githib aurora issue #3 via .
        (CG\*\*2) \* NENBW \*N   =  S2

        
        """
        lsd_factor1 = (1.0 / self.coherent_gain) * np.sqrt(
            1.0 / (self.nenbw * self.num_samples_window)
        )
        lsd_factor2 = 1.0 / np.sqrt(self.S2)
        if not np.isclose(lsd_factor1, lsd_factor2):
            print(f"factor1 {lsd_factor1} vs factor2 {lsd_factor2}")
            print("Incompatible spectral density factors")
            raise Exception

    @property
    def taper(self):
        return self._taper

    @taper.setter
    def taper(self, x):
        self._taper = x
        self._S1 = None
        self._S2 = None

    @property
    def apodization_factor(self):
        if self._apodization_factor is None:
            self._apodization_factor = np.sqrt(self.nenbw) * self.coherent_gain
        return self._apodization_factor
